from typing import List, Union

from unvibe.suite import TestCase as MyTestCase

from unvibe.magic import MagicFunction, MagicEntity, MagicClass
from unvibe.log import log

project_name = 'unvibe'
magic_entities: List[MagicEntity] = []
config = None
TestCase = MyTestCase


def reset():
    magic_entities.clear()


def ai(thing: Union[type, callable]) -> Union[MagicFunction, type]:
    """The @ai annotation used on functions and classes"""
    if isinstance(thing, type):
        mc = MagicClass(thing)
        log(f'Adding {mc} to magic classes')
        magic_entities.append(mc)
        return mc
    elif callable(thing):
        mf = MagicFunction(thing)
        magic_entities.append(mf)
        log(f'Adding {mf} to magic functions')
        return mf
    else:
        raise ValueError(f'@ai expected a class or a function, got {thing}')




import re
import traceback
import unittest
from copy import copy
from pprint import pprint
from random import random
from typing import List, Dict

from yaspin import yaspin

from unvibe import magic_entities
from unvibe.llm import ai_call
from unvibe.suite import UnvibeTestResult
from unvibe.tests_container import TestsContainer
from unvibe.config import config, config_get_or
from unvibe.log import log
from unvibe.magic import cleanup_implementation, MagicEntity
from unvibe.rand import up_to_1
from unvibe.state import State
from unvibe.ui import create_page_and_open_browser


def generate_new_state(count, state: State, temperature: float, tests_container: TestsContainer) -> State:
    """Generates a new state for program space search"""
    new_state = State()
    new_state.mes = state.mes
    new_state.count = count
    new_state.tests = state.tests
    new_state.temperature = temperature
    first_error = [state.errors[0]] if len(state.errors) > 0 else []
    prompt, resp_text = ai_call(state.mes, state.context, state.tests, first_error, temperature)
    new_state.prompt = prompt
    new_state.ai_output = resp_text
    impls = parse_ai_output(resp_text)
    new_state.impls = impls
    pprint(impls)
    # If returned implementations is subset of expected implementations:
    has_all_impls = {me.name for me in state.mes} <= set(impls.keys())

    if has_all_impls:
        log(f'Received {len(impls)} implementations, expected {len(state.mes)}')
        for me in new_state.mes:
            if me.name in impls:
                cleaned_up = cleanup_implementation(impls[me.name], me.__class__)
                me.set_impl(cleaned_up)
            else:
                raise Exception(f'Expected implementation for {me.name}')
        new_state.context = new_state.build_context_from_magic_entities()
        run_tests(tests_container, new_state)
        # Reset implementations:
        for me in new_state.mes:
            me.set_impl(None)
    else:
        log('LLM OUTPUT:', resp_text)
        new_state.score = 0
        expected_func_names = ', '.join([me.name for me in new_state.mes])
        new_state.errors = [f'Expected implementations for {expected_func_names}.\n'
                            f'Did you use one <implement> for each function? Try again.']
        log(new_state.errors)
    return new_state


def start_search(mes: List[MagicEntity], tests_container: TestsContainer, sources='', display_tree=False):
    log('Using model', config['ai']['model'])

    # Check all magic entities are registered
    for me in mes:
        assert me in magic_entities, f'{me} not registered with @ai'

    # Run the tree search
    root, states = search(mes, tests_container, sources, display_tree)
    best_state = states[0]
    if display_tree:
        file = create_page_and_open_browser(root)
        print(f'Created execution report {file}')
    return best_state


def get_temperatures(depth):
    random_spread = config_get_or('search', 'random_spread', 2)
    if depth == 0:
        random_spread = config_get_or('search', 'initial_spread', 10)
    random_type = config_get_or('search', 'random_type', 'uniform')
    max_temperature = config_get_or('search', 'max_temperature', 0.7)

    # Generate a bunch of rand temperatures, but always try temp=0
    if random_type == 'increasing':
        return [up_to_1.pop(0) for _ in range(random_spread)]
    elif random_type == 'uniform':
        return [0] + [random() * max_temperature for _ in range(random_spread)]
    else:
        raise Exception(f'Unknown random_type "{random_type}". Use either "increasing" or "uniform"')


def build_initial_context(mes, sources):
    if len(sources) > 0:
        # We have a list of source files, so we replace each MagicEntity with its implementation
        new_context = copy(sources)
        for me in mes:
            if me.impl is None:
                continue
            idx = sources.find(me.orig_code)
            if idx < 0:
                log(f'Could not find {me.orig_code} in context')
                continue
            else:
                new_context = new_context.replace(me.orig_code, me.impl)
        return new_context
    else:
        # We don't have files, just MEs, so we concatenate the implementations
        new_context = ''
        for me in mes:
            new_context += me.orig_code + '\n'
        return new_context


def search(mes: List[MagicEntity], test_container: TestsContainer, sources='', display_tree=False):
    take_best_n = config_get_or('search', 'take_best_n', 3)
    max_depth = config_get_or('search', 'max_depth', 10)

    # Generate the root state
    root = State()
    root.count = 0
    root.mes = mes
    root.temperature = 0.0
    root.tests = test_container.get_source()
    assert type(root.tests) == str
    root.errors = []
    root.score = -1  # root state has no score
    root.context = build_initial_context(mes, sources)
    root.orig_context = root.context
    states = [root]

    found = False
    count = 0
    spinner = yaspin()
    spinner.start()
    top_score = -1
    for depth in range(max_depth):
        log('Depth', depth)
        new_states = []
        # For each state, generate a bunch of new states feeding back the current test errors
        for state in states:
            # Generate a bunch of new states: generate code with LLMs and run the tests to get the score
            for temp in get_temperatures(depth):
                log('=============================')
                log('Temperature', temp)
                count += 1
                new_state = generate_new_state(count, state, temp, test_container)
                if new_state is None:
                    continue
                state.children.append(new_state)
                new_states.append(new_state)
                if new_state.score > top_score:
                    top_score = new_state.score
                    spinner.text = f'Depth {depth}, Top score: {top_score:.2f}'
                if new_state.score == 1:
                    # Early quit
                    log('Found perfect score')
                    found = True
                    break
            if found: break
        states = states + new_states
        states = sorted(states, key=lambda s: random())
        states.sort(key=lambda s: s.score, reverse=True)
        top_score = states[0].score
        log('Scores   ', [s for s in states], 'Picking the best', take_best_n)
        states = states[:take_best_n]
        # log('Selected ', [s for s in states])
        if display_tree:
            create_page_and_open_browser(root)
        if found: break
    spinner.stop()
    return root, states


def run_tests(tests_container: TestsContainer, new_state: State):
    """Runs the tests and saves the score/assertions info/errors in the new_state"""
    test_suite = tests_container.generate_test_suite()
    runner = unittest.TextTestRunner(resultclass=UnvibeTestResult)
    result: UnvibeTestResult = runner.run(test_suite)
    if result.testsRun == 0:
        raise Exception(f"Test class {test_suite} has no tests.")
    errors_count = len(result.failures) + len(result.errors)
    try:
        if not result.unvibe_test_case:
            raise Exception('Not using unvibe.TestCase')
        log('Using unvibe.TestCase')
        total_assertions = tests_container.count_assertions()
        log(f"Total assertions: {total_assertions}, Passed: {result.ass_passed}, "
            f"Executed: {result.ass_executed}, Failed: {result.ass_failed}")
        if total_assertions == 0:
            raise Exception('Invalid assertions count')
        score = result.ass_passed / max(total_assertions, result.ass_executed)
        tot_executed, tot_passed, tot_failed = result.ass_executed, result.ass_passed, result.ass_failed
    except Exception as exc:
        log(exc)
        score = 1 - errors_count / result.testsRun
        tot_passed, tot_executed, total_assertions, tot_failed = None, None, None, None
    log(f'Score: {score}')
    error_strings = []
    for test_suite, error_str in result.errors + result.failures:
        cleaned_up = cleanup_error_str(error_str)
        error_strings.append(cleaned_up)
        log(cleaned_up)
    new_state.passed_assertions = tot_passed
    new_state.executed_assertions = tot_executed
    new_state.failed_assertions = tot_failed
    new_state.total_assertions = total_assertions
    new_state.errors = error_strings
    new_state.score = score


def remove_lines_with(lines, is_target, minus=1, plus=2):
    try:
        target_line = -1
        for i, line in enumerate(lines):
            if is_target(line):
                target_line = i
                break
        if target_line >= 0:
            # Remove the target line and those before and after:
            lines = lines[:target_line - minus] + lines[target_line + plus:]
        return lines
    except Exception as e:
        traceback.print_exc()
        return lines


def cleanup_error_str(error_str):
    # Remove every reference to the file path:
    error_str = re.sub(r'File ".*", line', 'line', error_str)
    # Find the line with "eval" and "self.func_name":
    lines = error_str.split('\n')

    lines = remove_lines_with(lines, lambda line: 'eval(' in line and '{self.func_name}' in line)
    lines = remove_lines_with(lines, lambda line: 'exec(code)' in line)
    lines = remove_lines_with(lines, lambda line: 'return ___eval(' in line)
    lines = remove_lines_with(lines, lambda line: 'line ' in line and 'in wrapper' in line, minus=0, plus=2)
    lines = remove_lines_with(lines, lambda line: 'line ' in line and 'in wrapper' in line, minus=0, plus=2)

    # Put back together the redacted lines
    error_str = '\n'.join(lines)
    return error_str


def parse_ai_output(t: str) -> Dict:
    # find anything between <implements name="..."> and </implement>
    found = re.findall(r'<implement name="(.+?)">(.*?)</implement>', t, re.DOTALL)
    found_dict = dict(found)
    return found_dict


def index_of_first_non_empty_char(line: str) -> int:
    for i, c in enumerate(line):
        if c != ' ':
            return i
    return -1


def remove_extra_indentation(code: str):
    lines = code.split('\n')
    for line in lines:
        i = index_of_first_non_empty_char(line)
        if i >= 0:
            break
    ret = []
    for line in lines:
        ret.append(line[i:])
    return '\n'.join(ret)



import unittest
from collections import defaultdict

from unittest import TestSuite, TextTestResult, TestResult


def jet(obj, keys):
    keys = keys.split('.')
    try:
        x = getattr(obj, keys[0])
        for key in keys[1:]:
            x = getattr(x, key)
        return x
    except AttributeError:
        return None


class UnvibeTestResult(TestResult):
    ass_passed: int
    ass_executed: int
    ass_failed: int

    def __init__(self, stream, descriptions, verbosity):
        super().__init__(stream, descriptions, verbosity)
        self.failed_tests = defaultdict(lambda: 0)
        self.ass_passed = 0
        self.ass_executed = 0
        self.ass_failed = 0

    def addError(self, test, err):
        method_name = test._testMethodName
        self.failed_tests[method_name] += 1
        self.ass_failed += 1
        self.ass_executed += 1
        super().addError(test, err)


class CountingTestSuite(TestSuite):
    """A unittest.TestSuite that counts the number of successful assertions."""

    def __init__(self, tests=()):
        self.unvibe_test_case = False
        self.ass = dict()
        super().__init__(tests)

    def _tearDownPreviousClass(self, test, result):
        super()._tearDownPreviousClass(test, result)


class TestCase(unittest.TestCase):
    """
    Wraps all assertEqual, assertTrue etc. methods from unittest.TestCase,
    but counting the successful ones in a global counter
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name in dir(self):
            if name.startswith('assert'):
                assert_method = getattr(self, name)
                setattr(self, name, self._wrap_assert(assert_method))

    # def print_status(self, when, method):
    #     method_name = jet(self, '_testMethodName')
    #     tuple = (
    #         jet(self, '_outcome.result.ass_passed'),
    #         jet(self, '_outcome.result.ass_executed'),
    #         jet(self, '_outcome.result.ass_failed')
    #     )
    #     print(when, '\t\t', method_name, '\t\t', tuple)

    def _wrap_assert(self, assert_method):
        def wrapper(*args, **kwargs):
            # self.print_status('before', assert_method)
            self._outcome.result.unvibe_test_case = True
            try:
                assert_method(*args, **kwargs)
                self._outcome.result.ass_executed += 1
                self._outcome.result.ass_passed += 1
                # self.print_status('after', assert_method)
            except Exception as e:
                self._outcome.result.ass_executed += 1
                self._outcome.result.ass_failed += 1
                # self.print_status('after exc', assert_method)
                raise e

        return wrapper

    # def __setattr__(self, name, value):
    #     """standard implementation for setattr:"""
    #     print('SETATTR', name, value)
    #     self.__dict__[name] = value



from typing import Dict, List

from unvibe import MagicEntity


class State:
    orig_context: str  # all the files at the beginning of the search
    context: str  # the orig_context concatenated with the current implementations from this state
    prompt: str
    ai_output: str
    impls: Dict[str, str]  # For each function, the source-code implementation: func_name -> code implementation.
    mes: List[MagicEntity]  # the magic functions with their original definitions and "current" implementations
    tests: str  # the source code of the test class
    errors: List[str]  # the errors we got from the tests
    score: float  # the score for this solution
    passed_assertions: int
    total_assertions: int
    executed_assertions: int
    failed_assertions: int
    children: List['State']
    temperature: float = None
    count = None

    def __init__(self):
        self.orig_context: str = ''
        self.context: str = None
        self.prompt: str = None
        self.ai_output: str = None
        self.impls: Dict[str, str] = {}
        self.mes: List[MagicEntity] = None
        self.tests: str = None
        self.errors: List[str] = []
        self.score: float = None
        self.passed_assertions: int = 0
        self.total_assertions: int = 0
        self.executed_assertions: int = 0
        self.failed_assertions: int = 0
        self.children: List['State'] = []
        self.temperature: float = None
        self.count = None

    def build_context_from_magic_entities(self):
        context = self.orig_context
        for mf in self.mes:
            context += (mf.impl if mf.impl is not None else mf.clean_orig_code) + '\n'
        return context

    def __repr__(self):
        return f'State(#{self.count}, score={self.score:.2f}, errors={len(self.errors)}, temp={self.temperature:.3f})'

    def to_dict(self):
        return {
            'count': self.count,
            'context': self.context,
            'prompt': self.prompt,
            'ai_output': self.ai_output,
            'impls': self.impls,
            'mfs': [mf.to_dict() for mf in self.mes],
            'tests': self.tests,
            'errors': self.errors,
            'score': self.score,
            'passed_assertions': self.passed_assertions,
            'total_assertions': self.total_assertions,
            'executed_assertions': self.executed_assertions,
            'failed_assertions': self.failed_assertions,
            'temperature': self.temperature,
            'children': [child.to_dict() for child in self.children]
        }



from http.server import HTTPServer, SimpleHTTPRequestHandler
from os import path

from unvibe.log import log

PORT = 8000
DIRECTORY = path.join(path.dirname(__file__), 'html')

log('Starting server on port', PORT, 'and serving files from', DIRECTORY)
httpd = HTTPServer(('localhost', PORT), SimpleHTTPRequestHandler)
httpd.serve_forever()



import os
import sys

try:
    import tomllib
except ModuleNotFoundError:
    import pip._vendor.tomli as tomllib  # For Python < 3.11

config_file_name = '.unvibe.toml'
doc_url = 'https://github.com/santinic/unvibe'


def read_config():
    if os.getenv('UNITAI_CONFIG'):
        file_path = os.getenv('UNITAI_CONFIG')
    else:
        file_path = os.path.join('.', config_file_name)
    if not os.path.exists(file_path):
        print(f'No {config_file_name} file found in {file_path}')
        print(f'Please check documentation: {doc_url}')
        sys.exit(1)
    print(f'Reading {file_path} config')
    with open(file_path, 'r') as file:
        config = tomllib.loads(file.read())
    check_config(config)
    return config


def check_config(config):
    if 'ai' not in config:
        print(f'Section [ai] not found in {config_file_name}. Please check documentation: {doc_url}')
        sys.exit(1)
    if 'provider' not in config['ai']:
        print(f'Key "provider" not found in [ai] section of {config_file_name}. Please check documentation: {doc_url}')
        sys.exit(1)
    if 'model' not in config['ai']:
        print(f'Key "model" not found in [ai] section of {config_file_name}. Please check documentation: {doc_url}')
        sys.exit(1)


def config_get_or(section, key, default=None):
    if section not in config:
        raise Exception(f'Section [{section}] not found in {config_file_name}. Please check documentation: {doc_url}')
    if key in config[section]:
        return config[section][key]
    return default


config = read_config()



import os
import pickle

from unvibe.config import config_get_or

cache_file = 'unvibe_cache.pkl'


def disk_cached(func):
    if config_get_or('search', 'cache', True) is False:
        return func

    def wrapper(*args, **kwargs):
        key = f"{func.__name__}_{args}_{kwargs}"

        # Create cache file if it doesn't exist
        if not os.path.exists(cache_file):
            with open(cache_file, 'wb') as f:
                pickle.dump({}, f)

        # Read existing cache
        with open(cache_file, 'rb') as f:
            cache = pickle.load(f)

        # Check if result is cached
        if key in cache:
            return cache[key]

        # Compute result
        result = func(*args, **kwargs)

        # Store result in cache
        cache[key] = result
        with open(cache_file, 'wb') as f:
            pickle.dump(cache, f)

        return result

    return wrapper


def reset_cache():
    if os.path.exists(cache_file):
        os.remove(cache_file)


def get_keys():
    if not os.path.exists(cache_file):
        return set()

    with open(cache_file, 'rb') as f:
        cache = pickle.load(f)

    return set(cache.keys())



import re
import sys
import inspect
import traceback
from typing import Union, Type

from termcolor import colored

annotation_text = '@ai'

impl_is_none_msg = colored(
    'You are probably running a function with @ai annotation without using Unvibe as runner.\n'
    'To run Unvibe you need to define unit-tests and annotate functions with @ai and then run:\n', 'red')
impl_is_none_msg += (
    '$ unvibe src/ tests/                     # To discover all tests in the tests/\n'
    '$ unvibe src/main.py tests/test_main.py  # To run only the tests in the indicated file')


def as_short_code(code, max_len=150):
    short_code = code.replace('\n', ' ').strip()[:max_len] + '…'
    short_code = re.sub(r'\s+', ' ', short_code)
    return short_code


def split_imports_and_code(impl):
    imports = []
    lines = impl.split('\n')
    for i, line in enumerate(lines):
        clean_line = line.strip()
        if line == '':
            continue
        elif clean_line.startswith('import ') or clean_line.startswith('from '):
            imports.append(line)
        else:
            break
    code = '\n'.join(lines[i:])
    imports = '\n'.join(imports)
    return imports, code


def cleanup_implementation(code, cls: Union[Type['MagicFunction'], Type['MagicClass']]):
    code = remove_indentation(code, cls)
    code = remove_annotation(code)
    return code


def remove_indentation(code, cls):
    lines = code.split('\n')
    indent = ''
    first_def = 'class ' if cls == MagicClass else 'def '
    for line in lines:
        if line.strip().startswith(first_def):
            indent = line.split(first_def)[0]
            break
    ret = []
    for line in lines:
        ret.append(line.removeprefix(indent))
    return '\n'.join(ret)


def remove_annotation(code):
    lines = code.split('\n')
    found = None
    for i, line in enumerate(lines):
        if line.strip() == annotation_text:
            found = i
    if found:
        lines.pop(found)
        return '\n'.join(lines)
    return code


class MagicEntity:
    orig: str
    clean_orig_code: str

    def __init__(self, orig):
        self.orig = orig
        self.orig_code = inspect.getsource(orig)
        self.clean_orig_code = self.orig_code.replace(annotation_text, '')  # remove @ai
        self.name = orig.__name__
        self.impl = None

    def set_impl(self, impl):
        self.impl = impl

    def to_dict(self):
        return {
            'name': self.name,
            'orig_code': self.orig_code,
            'impl': self.impl
        }


class MagicFunction(MagicEntity):
    """
    This is a wrapper for the functions annotated with @ai.
    You can swap the implementation of the MagicFunction and let the Unit-Tests invoke
    the original function, but actually eval the implementation.
    The search algorithm will create many different implementations and invoke
    magic_function.set_impl(code) to set the implementation.
    """

    def __repr__(self):
        short_code = as_short_code(self.orig_code)
        return f'MagicFunction({short_code}...)'

    def __call__(self, *args, **kwargs):
        if self.impl is None:
            raise Exception(f'Implementation not set for {self}.\n\n{impl_is_none_msg}')

        imports, code = split_imports_and_code(self.impl)
        ___eval = eval
        try:
            exec(imports, globals())  # run the imports
            exec(code, globals())  # define the function
        except IndentationError as exc:
            traceback.print_exc()
            print('IndentationError produced with:')
            print('Imports:', imports)
            print('Code:', code)
            sys.exit(1)
        return ___eval(f'{self.name}(*args, **kwargs)')  # then call it


class MagicClass(MagicFunction):
    def __repr__(self):
        short_code = as_short_code(self.orig_code)
        return f'MagicClass({short_code}...)'



import importlib
import inspect
import sys
import unittest
from abc import abstractmethod
from pathlib import Path
from unittest import TestSuite, TestLoader, TestCase

from unvibe.suite import CountingTestSuite


def count_assertions(src: str) -> int:
    lines = src.split('\n')
    count = 0
    for line in lines:
        if 'self.assert' in line and not line.strip().startswith('#'):
            count += 1
    return count

def import_from_path(module_name, file_path):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


class TestsContainer:
    def __init__(self):
        pass

    @abstractmethod
    def generate_test_suite(self) -> TestSuite:
        pass

    @abstractmethod
    def get_source(self) -> str:
        pass

    @abstractmethod
    def count_assertions(self):
        pass


class ClassTestsContainer(TestsContainer):
    def __init__(self, test_suite: TestSuite):
        self.test_suite = test_suite

    def generate_test_suite(self) -> TestSuite:
        loader = TestLoader()
        loader.suiteClass = CountingTestSuite
        test_suite = loader.loadTestsFromTestCase(self.test_suite)
        return test_suite

    def get_source(self):
        return inspect.getsource(self.test_suite)

    def count_assertions(self):
        src = inspect.getsource(self.test_suite)
        return count_assertions(src)


class FolderPatternTestsContainer(TestsContainer):
    def __init__(self, test_folder, pattern):
        super().__init__()
        self.test_folder = test_folder
        self.pattern = pattern
        self.test_sources = self._get_all_test_sources()

    def _get_all_test_sources(self) -> str:
        tests_path = Path(self.test_folder)
        if tests_path.is_file():
            tests_file_names = [tests_path]
        else:
            tests_file_names = list(tests_path.glob(self.pattern))
            if len(tests_file_names) == 0:
                print('No test files found in', self.test_folder)
                sys.exit(1)
            print(f'{len(tests_file_names)} test files found')
        ctx = '\n'.join([f.read_text() for f in tests_file_names])
        return ctx

    def generate_test_suite(self):
        if Path(self.test_folder).is_file():
            module = import_from_path('module', self.test_folder)
            return unittest.TestLoader().loadTestsFromModule(module)
        else:
            # is a folder
            test_loader = unittest.TestLoader()
            test_loader.suiteClass = CountingTestSuite  # Count the passed assertions if possible
            test_suite: TestSuite = test_loader.discover(self.test_folder, self.pattern)
            print(f'{test_suite.countTestCases()} test cases found')
            return test_suite

    def get_source(self):
        return self.test_sources

    def count_assertions(self):
        return count_assertions(self.test_sources)
